from utils import *

class Memory(object):
    '''A Critter's memory is responsible for relating the network to decisions.'''

    def __init__(self, critter, stochastic=True, learning=False):
        self.critter = critter
        self.network = critter.network
        self.stochastic = stochastic
        self.learning = learning
        self.competitive = critter.competitive
        self.rules = []
        self.set_biases()
        self.n_acts = len(critter.acts)
        self.exploitation = self.critter.world.features.get('EXPLOITATION', 1.0)

    def sensation2input(self, sensation):
        '''Convert a sensation string to an input index.'''
        return self.critter.sensible.index(sensation)

    def action2output(self, action):
        '''Convert an action to output index.'''
        return self.critter.acts.index(action)

    def output2action(self, output_index):
        '''Convert an output index to an action.'''
        return self.critter.acts[output_index]

    def decide(self):
        '''Choose an action: the index of an output unit of the network.'''
        if self.competitive:
            return random.randint(0, self.n_acts - 1)
        elif self.stochastic:
            if self.learning:
                return exp_luce_choice(self.network.layers[-1].activations, self.exploitation)
#                                       self.exploitation * self.critter.age)
            else:
                return luce_choice(self.network.layers[-1].activations)
        else:
            return argmax_index(self.network.layers[-1].activations, lambda x: x)

    # EXPLICIT RULES RELATING SENSORY INPUT AND ACTION OUTPUT

    def add_rule(self, rule):
        '''Add a rule and assign weights to the network for it.'''
        self.rules.append(rule)
        self.rule2weights(rule)

    def rule_feature(self, rule):
        '''Feature of a rule.'''
        return rule[0]

    def rule_action(self, rule):
        '''Action of a rule.'''
        return rule[1]

    def rule_actions(self, rule):
        '''List of action of a rule.'''
        return rule[1:]

    def rule2weights(self, rule):
        '''Assign weights to the network for a rule.'''
        inp = self.sensation2input(self.rule_feature(rule))
        for act_weight in self.rule_actions(rule):
            act = self.action2output(act_weight[0])
            weight = act_weight[1]
            self.network.layers[-1].weights[act][inp] = weight

    def dflt2bias(self, act, bias):
        '''Convert a default action to a network bias.'''
        act_index = self.action2output(act)
        self.network.layers[-1].set_bias(act_index, bias)

    def set_biases(self):
        '''Set biases for the network.'''
        if self.network.has_bias:
            for act_bias in self.critter.act_bias:
                self.dflt2bias(act_bias[0], act_bias[1])
